#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>
#include <pwd.h>
#include <time.h>
#include <limits.h>
#include "proc_parser.h"
#include "human.h"

static char c_flag;

static char *get_cmdline(const pid_t pid) {
	static char path[PATH_MAX + 1];
	snprintf(path, sizeof(path), "/proc/%d/cmdline", pid);

	int fd = open(path, O_RDONLY);
	if (fd < 0)
		return NULL;

	ssize_t size = read(fd, path, sizeof(path));
	close(fd);

	if (size == 0)
		return NULL;

	for (ssize_t i = 0; i < size; i++) {
		if (path[i] == '\0' && i != size - 1)
			path[i] = ' ';
	}

	return path;
}

static int pscan(const pid_t pid) {
	struct mu_proc proc;
	if (mu_proc_parser("ps", pid, &proc))
		return 1;

	/* Uid */
	char *name = "unknow";
	struct passwd *pw = getpwuid(proc.uid);
	if (pw != NULL)
		name = pw->pw_name;

	/* Time */
	unsigned int rtime = (proc.utime + proc.stime) / sysconf(_SC_CLK_TCK);

	/* Print */
	char virt[MU_HUMAN_BUF_SIZE + 1];
	strcpy(virt, mu_humansize(proc.vsize, 1024));

	char rss[MU_HUMAN_BUF_SIZE + 1];
	strcpy(rss, mu_humansize(proc.vmrss * 1024, 1024));

	char *prog = (c_flag) ? get_cmdline(pid) : proc.cmdline;
	if (prog == NULL)
		prog = proc.cmdline;

	printf("%6d %8s %4ld %4ld %8s %8s %2c %02um:%02us %2s\n", proc.pid, name, proc.priority, proc.nice, virt, rss, proc.state, rtime / 60, rtime % 60, prog);
	return 0;
}

int main(int argc, char **argv) {
	int opt;
	while ((opt = getopt(argc, argv, "c")) != -1) {
		switch (opt) {
			case 'c':
				c_flag = 1;
				break;

			default:
				puts("ps [c] [PID]\n\t-c Print cmdline instead of program name");
				return 0;
		}
	}

	argv += optind;
	argc -= optind;

	puts("   PID     USER  PRI NICE     VIRT      RSS  S RTIME   CMD");

	int ret = 0;
	if (argc == 0) {
		DIR *dp = opendir("/proc");
		if (dp == NULL) {
			fprintf(stderr, "ps: /proc: %s\n", strerror(errno));
			return 1;
		}

		struct dirent *ep;
		while ((ep = readdir(dp)) != NULL) {
			pid_t pid = strtoul(ep->d_name, 0L, 10);
			if (pid)
				if (pscan(pid))
					ret = 1;
		}

		closedir(dp);
	}

	else {
		for (int i = 0; i < argc; i++) {
			pid_t pid = strtoul(argv[i], 0L, 10);
			if (pid)
				if (pscan(pid))
					ret = 1;
		}
	}

	return ret;
}
